from freezegun import freeze_time
from datetime import datetime, timedelta

from steinloss.loss_calculator import Loss_Calculator, TimeTable
from steinloss.packet_entity import sent_package, receive_package


class TestLossCalculator:
    def test_initiate_loss_calculator(self):
        Loss_Calculator()

    def test_loss_calculator_can_add_packet_entry(self, freezer):
        loss_calculator = Loss_Calculator()
        packet = sent_package('1', datetime.now())

        loss_calculator.add(packet)

        assert packet in loss_calculator

    @freeze_time("00:00:00", auto_tick_seconds=1)
    def test_two_entries_with_more_than_a_second_apart_should_not_be_in_same_key(self, freezer):
        loss_calculator = Loss_Calculator()
        first_time = datetime.now()
        second_time = datetime.now()

        packet_one = sent_package('1', first_time)
        packet_two = sent_package('2', second_time)
        loss_calculator.add(packet_one)
        loss_calculator.add(packet_two)

        assert loss_calculator[first_time].sent == 1
        assert loss_calculator[second_time].sent == 1

    @freeze_time("00:00:00", auto_tick_seconds=0.5)
    def test_loss_calculator_can_add_two_entries_at_same_timestamp(self, freezer):
        loss_calculator = Loss_Calculator()
        at_zero_second = datetime.now()
        at_zero_point_5_second = datetime.now()
        packet_at_zero_second = sent_package('1', at_zero_second)
        packet_at_zero_point_5_second = sent_package('2', at_zero_point_5_second)

        loss_calculator.add(packet_at_zero_second)
        loss_calculator.add(packet_at_zero_point_5_second)

        assert loss_calculator[at_zero_second].sent == 2

    def test_loss_calculator_counts_sent_packages_at_timestamp(self):
        loss_calculator = Loss_Calculator()
        time_stamp = datetime.now()

        loss_calculator.add(sent_package('1', time_stamp))
        loss_calculator.add(sent_package('2', time_stamp))

        assert loss_calculator[time_stamp].sent == 2

    def test_loss_calculator_counts_received_packages_at_timestamp(self):
        loss_calculator = Loss_Calculator()
        time_stamp = datetime.now()

        loss_calculator.add(sent_package('1', time_stamp))
        loss_calculator.add(receive_package('1', '1', time_stamp))

        assert loss_calculator[time_stamp].received == 1

    def test_loss_calculator_counts_loss_when_packet_do_not_return(self):
        loss_calculator = Loss_Calculator()
        time_stamp = datetime.now()

        loss_calculator.add(sent_package('1', time_stamp))
        loss_calculator.add(sent_package('2', time_stamp))
        loss_calculator.add(sent_package('3', time_stamp))

        loss_calculator.add(receive_package('1', '1', time_stamp))
        loss_calculator.add(receive_package('3', '2', time_stamp))

        assert loss_calculator[time_stamp].loss == 1

    def test_sent_package_added_to_calculator_is_logged(self):
        loss_calculator = Loss_Calculator()
        time_stamp = datetime.now()
        package = sent_package('1', time_stamp)

        loss_calculator.add(package)

        assert loss_calculator[package.id].is_sent()

    def test_sent_package_added_to_calculator_is_not_received(self):
        loss_calculator = Loss_Calculator()
        time_stamp = datetime.now()
        package = sent_package('1', time_stamp)

        loss_calculator.add(package)

        assert not loss_calculator[package.id].is_received()

    @freeze_time("00:00:00")
    def test_convert_datetime_max_value_for_rounding_down(self):
        time_table = TimeTable()

        # 1000 milliseconds = 1 second
        at_999_milliseconds = datetime.now() + timedelta(milliseconds=999)
        at_0_second_string = time_table.convert_time_to_key(at_999_milliseconds)

        assert at_0_second_string == "00:00:00"

    @freeze_time("00:00:00")
    def test_convert_datetime_returns_seconds_in_string(self):
        time_table = TimeTable()

        # 1000 milliseconds = 1 second
        at_1_milliseconds = datetime.now() + timedelta(seconds=1)
        at_1_second_string = time_table.convert_time_to_key(at_1_milliseconds)

        assert at_1_second_string == "00:00:01"

    def test_get_last_package(self):
        loss_calculator = Loss_Calculator()

        package = sent_package("1")
        loss_calculator.add(package)

        last_package = loss_calculator.get_last_packages(1).pop()
        assert last_package.sent_at == package.time

    def test_get_n_latest_packages_where_n_is_equal_to_length(self):
        loss_calculator = Loss_Calculator()
        loss_calculator.add(sent_package("1"))
        loss_calculator.add(sent_package("2"))

        packages = loss_calculator.get_last_packages(2)

        assert len(packages) == 2

    @freeze_time(auto_tick_seconds=1)
    def test_get_n_latest_packages(self):
        loss_calculator = Loss_Calculator()
        p1 = sent_package("1", datetime.now())
        p2 = sent_package("2")
        p3 = sent_package("3")

        loss_calculator.add(p1)
        loss_calculator.add(p2)
        loss_calculator.add(p3)

        packages = loss_calculator.get_last_packages(2)

        assert not any(pack.sent_at == p1.time for pack in packages)
        assert any(pack.sent_at == p2.time for pack in packages)
        assert any(pack.sent_at == p3.time for pack in packages)

    # Test ideas:
    # what happens after 24H



import asyncio
import pytest

from datetime import datetime
from steinloss.Data_Presenter import Data_Presenter
from steinloss.packet_entity import sent_package, receive_package
from steinloss.server import Server

kilobyte = 1024


class TestServer:

    def teardown_method(self):
        Data_Presenter.clear_instance()

    @pytest.mark.parametrize(
        "speed, interval", [(kilobyte, 1), (2 * kilobyte, 0.5), (4 * kilobyte, 0.25)]
    )
    def test_two_kilobytes_should_set_interval_to_half(self, speed, interval):
        server = Server()
        server.speed = speed

        assert server.interval == interval

    def test_construct_server_with_five_kilobytes(self):
        five_kilobytes = 5120
        server = Server(five_kilobytes)

        assert server.interval == 0.2

    @pytest.mark.parametrize("packets", [1, 2, 3, 4])
    def test_send_packet_should_save_entry(self, mocker, packets):
        mocker.patch('socket.socket')
        time = datetime.now()
        mocker.patch('steinloss.server.Server.timestamp', return_value=time)
        server = Server()

        for i in range(packets):
            server.send_packet('ip_address')

        assert server.data_presenter.get_time_table()[time].sent == packets

    def test_wait_for_probe_should_return_address_of_probe(self, mocker):
        server = Server()
        fake_address = '0,1,2,3', 4321
        mocker.patch('socket.socket.recvfrom', return_value=[b"package", fake_address])

        assert server.wait_for_probe() == fake_address

    @pytest.mark.parametrize(
        "iterations", [1, 2, 3]
    )
    def test_send_packet_should_should_send_id_to_client(self, mocker, iterations):
        server = Server()
        probe_address = '1,3,3,7', 1001
        mocked_sendto = mocker.patch('socket.socket.sendto')
        for i in range(iterations):
            server.send_packet(probe_address)
            mocked_sendto.assert_called_with(('%d' % i).encode(), probe_address)

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "speed, duration, packages", [(4096, 0.24, 1), (8192, 0.24, 2)]
    )
    async def test_serve_forever_should_send_packets_according_to_speed(
            self, mocker, event_loop, speed, duration, packages):
        mocked_sendto = mocker.patch('steinloss.server.Server.send_packet')

        server = Server(speed=speed)

        event_loop.create_task(server.serve_forever('fake_address'))
        await asyncio.sleep(duration, loop=event_loop)

        assert mocked_sendto.call_count == packages

    # shutdown should log rest, and close socket
    def test_calculate_packet_loss_should_answer_in_pct_of_lost_packets(self):
        server = Server()
        time = datetime.now()

        server.save_entry(sent_package("1", time))
        server.save_entry(sent_package("2", time))
        server.save_entry(receive_package("1", "1", time))

        assert server.calculate_packet_loss_in_pct(time) == 0.5


# divide by zero in calculate loss


class FakeTime:
    def __init__(self):
        self.__time = 0

    def time(self):
        return self.__time

    def sleep(self, seconds):
        self.__time += seconds



from argparse import ArgumentParser

import pytest

from steinloss.stein_parser import setup


class TestParser:

    def setup_method(self):
        std_parser = ArgumentParser()
        self.parser = setup(std_parser)

    def test_stein_parser_stores_server_variable(self):
        args = ['--server']

        parser = self.parser.parse_args(args)

        assert parser.server

    def test_stein_parser_server_abbreviation_returns_true(self):
        args = ['-s']

        self.parser.print_help()
        parser = self.parser.parse_args(args)

        assert parser.server

    def test_stein_parser_probe_variable(self):
        args = "--probe 0.0.0.0".split(' ')

        parser = self.parser.parse_args(args)

        assert parser.probe

    def test_stein_parser_probe_abbreviation_returns_true(self):
        args = '-p 0.0.0.0'.split(' ')

        parser = self.parser.parse_args(args)

        assert parser.probe

    def test_raises_exit_when_neither_server_or_probe_is_selected(self):
        args = []

        with pytest.raises(SystemExit):
            parser = self.parser.parse_args(args)

    def test_default_port_is_set(self):
        args = ['-s']

        parser = self.parser.parse_args(args)

        assert isinstance(parser.port, int)

    def test_probe_exits_if_no_ip_is_provided(self):
        args = ['-p']

        with pytest.raises(SystemExit):
            parser = self.parser.parse_args(args)



from socket import AF_INET, SOCK_DGRAM
import socket


class Probe:
    packet_size = 1024

    def __init__(self, server_address, address=('', 7071)):
        self.window_size = 3
        self.lost = 0
        self.duplicate = 0
        self.reorder = 0
        self.sequence_number = None
        self.server_to_client_loss = 0
        self.suspect = [0] * 5
        self.server_address = server_address
        self.address = address
        self.id = 0
        self.sock = socket.socket(SOCK_DGRAM, AF_INET)
        self.sock.bind(self.address)

    def main(self):
        self.ping_server()
        self.run_loop()

    def run(self):
        try:
            self.respond_to_server('go')
            self.run_loop()
        except KeyboardInterrupt:
            pass
        finally:
            self.shutdown()

    def receive_packet(self):
        packet = self.sock.recv(self.packet_size).decode()

        self.consume_packet(packet)

        self.sequence_number = int(packet)

        self.respond_to_server(packet + f"_{self.id}")
        self.id += 1

        if self.lost == 0 or self.sequence_number == 0:
            lost_pct = 0
        else:
            lost_pct = self.lost / self.sequence_number
        print(
            f"received message: {self.sequence_number} | probe id: {self.id} | ",
            "server->probe loss: {:.2f}%".format(lost_pct * 100),
            end='\r')

    def respond_to_server(self, packet: str):
        self.sock.sendto(packet.encode(), self.server_address)

    def ping_server(self):
        self.respond_to_server('begin')

    def shutdown(self):
        pass

    def run_loop(self):
        while True:
            self.receive_packet()

    def consume_packet(self, packet):
        full_window = self.window_size + 2
        packet = int(packet)
        pos = packet % full_window

        if self.sequence_number is None:
            self.sequence_number = packet - 1

        if packet > self.sequence_number:  # det er en ny pakke
            # Move window forward from 2 being middle, to 3 being middle
            # [0,1,2,3,4]
            # [0,0,0,1,1]

            # [5,1,2,3,4]
            # [1,0,0,0,1]

            # kun kig på det her, hvis vi er hoppet two tal
            two_up = (pos - 2) % 5
            if self.suspect[two_up] == 1:
                self.lost += 1
            self.suspect[two_up] = 1

            if packet > (self.sequence_number + 1):
                one_up = (pos - 1) % 5
                if self.suspect[one_up] == 1:
                    self.lost += 1
                self.suspect[one_up] = 1

            # set my pos to found
            self.suspect[pos] = 0  # no loss
            pass

        if packet == self.sequence_number:
            # duplication
            pass
        if packet < self.sequence_number:
            # is it in reordering window?
            if packet < self.sequence_number - self.window_size:
                # it is an old package.
                # The loss of this package is already counted
                pass
            else:
                # in reordering window
                # Tell that one of the packages in the reordering window is found
                self.suspect[pos] = 0
                # update my pos to found
                pass

        self.sequence_number = max(packet, self.sequence_number)



import asyncio
import socket
from datetime import datetime, timedelta
from typing import Tuple
import time

from steinloss.Data_Presenter import Data_Presenter
from steinloss.packet_entity import sent_package, receive_package, Packet_entity

ONE_SECOND = 1

kilobyte = 1024
megabyte = 1024 * kilobyte
gigabyte = 1048576 * kilobyte


class Server:
    packet_size = kilobyte

    def __init__(self, speed=megabyte * 100, listening_address=('0.0.0.0', 7070),
                 runtime_of_test=ONE_SECOND * 60 * 30):
        self.last_sent_packet = 0
        self.last_received_packet = 0
        self.time_of_sample_size = runtime_of_test
        self.socket_timeout = 10  # seconds
        self.log_interval = 1
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.listening_address = listening_address
        self.id = 0
        self.__interval = 1
        self.speed = speed
        self.data_presenter = Data_Presenter.get_instance()

    @property
    def speed(self):
        return 1 / self.__interval * self.packet_size

    @speed.setter
    def speed(self, value):
        self.__interval = self.packet_size / value

    @property
    def interval(self):
        return self.__interval

    def send_packet(self, address):
        packet = "%d" % self.id

        package = sent_package(packet, self.timestamp())
        self.save_entry(package)
        self.id += 1
        self.server_socket.sendto(packet.encode(), address)

    def save_entry(self, package: Packet_entity):
        self.data_presenter.append(package)

    def run(self):
        self.ready_event_loop()
        self.ready_socket()

        try:
            address = self.wait_for_probe()
            # try:
            self.run_loop(address)
        except Exception as e:
            print("Server timeout. Client didn't connect to server")
            print(e)

    def run_loop(self, address):

        # loop part
        loop = asyncio.get_event_loop()

        listen_task = loop.create_datagram_endpoint(
            lambda: EchoServerProtocol(self),
            sock=self.server_socket
        )
        transport, protocol = loop.run_until_complete(listen_task)

        loop.create_task(self.log_forever())
        loop.create_task(self.serve_forever(address))
        # Running part
        print('loop is running')

        try:
            loop.run_until_complete(asyncio.sleep(self.time_of_sample_size))
            print("Test is complete")
        except KeyboardInterrupt:
            print("test got interrupted")
        finally:
            transport.close()
            self.shutdown()

    def wait_for_probe(self):
        self.server_socket.settimeout(self.socket_timeout)
        print("Server ready at: %s %s" % self.listening_address)
        request_and_address = self.server_socket.recvfrom(1024)

        address: Tuple[str, int] = request_and_address[1]
        print(f"Request from {address[0]}{address[1]}")
        return address

    def ready_socket(self):
        try:
            # Assign IP address and port number to socket
            self.server_socket.bind(self.listening_address)
        except socket.error as error:
            self.shutdown()

    def log(self):
        one_second_in_the_past = datetime.now() - timedelta(seconds=2)

        packet_loss = self.calculate_packet_loss_in_pct(one_second_in_the_past)

        sent = self.data_presenter.get_time_table()[one_second_in_the_past].sent
        received = self.data_presenter.get_time_table()[one_second_in_the_past].received

        print(f"{sent} packets sent last second |"
              + f" {received} packets received last second ",
              " | package loss: {:.2f}".format(packet_loss * 100),
              end='\r')

    async def log_forever(self):
        while True:
            self.log()
            await asyncio.sleep(self.log_interval)

    async def serve_forever(self, address):
        start_time = time.time()
        while True:
            self.send_packet(address)

            await asyncio.sleep(self.__interval - (time.time() - start_time) % self.__interval)

    def shutdown(self):
        self.server_socket.close()

        tasks = asyncio.Task.all_tasks()
        for task in tasks:
            task.cancel()

        loop = asyncio.get_event_loop()
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()

    def calculate_packet_loss_in_pct(self, timestamp: datetime):
        time_entry = self.data_presenter.get_time_table()[timestamp]
        packages_sent = time_entry.sent
        packages_recv = time_entry.received

        if packages_sent == 0 or packages_recv == 0:
            return 0
        else:
            return 1 - packages_recv / packages_sent

    def timestamp(self):
        return datetime.now()

    def ready_event_loop(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)


class EchoServerProtocol(asyncio.DatagramProtocol):
    def __init__(self, server):
        self.server = server
        self.transport = None

    def connection_made(self, transport):
        self.server.server_socket = transport
        self.transport = transport

    def datagram_received(self, data, addr):
        message = data.decode()
        numbers = message.split('_')
        sent_packet = numbers[0]
        received_packet = numbers[1]

        package = receive_package(sent_packet, received_packet, datetime.now())
        self.server.data_presenter.append(package)



from argparse import ArgumentParser


def setup(parser: ArgumentParser) -> ArgumentParser:
    parser.prog = "steinloss"
    parser.description = \
        "A tool for measuring a package loss, between two endpoints. The way it works, is by spinning up a server " \
        "endpoint, that waits for a incoming connection. When the server-side gets pinged by a probe, it will start " \
        "sending packages to it, and the probe will respond on each packages." \
        "The package loss is calculated by keeping track of the id of the packet." \
        "Data is shown on the a website on port 8080, on the server side"
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-s", "--server", action="store_true",
                       help="Determines if you're on the server side")
    group.add_argument("-p", "--probe", nargs=1, metavar='target_ip_address',
                       help="Determines if you're on the probe side. "
                            "You have to specify a ip address for the probe to target")

    parser.add_argument("-P", "--port", type=int, default=9090,
                        help="Which port to use. Have to be the same, as the servers port")

    return parser


def cli():
    parser = setup(ArgumentParser())
    parser.parse_args()



import asyncio
from pathlib import Path
import socket


def get_project_root() -> Path:
    """Returns project root folder."""
    return Path(__file__).parent.parent


async def repeat(func, interval=None, *args, **kwargs):
    """Run func every interval seconds.

    If func has not finished before *interval*, will run again
    immediately when the previous iteration finished.

    *args and **kwargs are passed as the arguments to func.
    """
    if interval is None:
        while True:
            await asyncio.gather(func(*args, **kwargs))
    else:
        while True:
            await asyncio.gather(
                func(*args, **kwargs),
                asyncio.sleep(interval),
            )


if __name__ == '__main__':
    UDP_IP = "127.0.0.1"
    UDP_PORT = 7070
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.bind((UDP_IP, UDP_PORT))

    sock.sendto('go'.encode(), (UDP_IP, 7071))

    id = 0
    while True:
        data, addr = sock.recvfrom(1024)  # buffer size is 1024 bytes
        print("received message: %s" % data.decode(), end='\r')
        data = data.decode() + f"_{id}"
        id += 1
        sock.sendto(data.encode(), (UDP_IP, 7071))



from datetime import datetime

import pytest
from freezegun import freeze_time

from steinloss.Data_Presenter import Data_Presenter
from steinloss.packet_entity import sent_package, receive_package


class TestDataPresenter:
    def teardown_method(self):
        Data_Presenter.clear_instance()

    def test_data_presenter_can_be_instantiated(self):
        my_singleton = Data_Presenter()
        assert my_singleton is not None

    def test_data_presenter_can_take_an_argument(self):
        test_data = sent_package("1")
        data_presenter = Data_Presenter.get_instance()

        data_presenter.append(test_data)

        assert data_presenter.latest_packages().pop().sent_at == test_data.time

    def test_data_presenter_instance_is_stored_as_a_class_variable(self):
        data_presenter = Data_Presenter()

        assert Data_Presenter.get_instance() is not None

    def test_data_presenter_cannot_be_initiated_twice(self):
        data_presenter = Data_Presenter()

        with pytest.raises(RuntimeError) as re:
            Data_Presenter()

        assert re

    def test_get_instance_initiates_the_class_if_not_initiated(self):
        data_presenter = Data_Presenter.get_instance()

        assert data_presenter is not None

    def test_data_from_different_get_instance(self):
        data_presenter1 = Data_Presenter.get_instance()
        data_presenter2 = Data_Presenter.get_instance()

        test_data = sent_package("1")
        data_presenter1.append(test_data)

        from_different_object = data_presenter2.latest_packages().pop()

        assert from_different_object.sent_at == test_data.time

    @freeze_time(auto_tick_seconds=1)
    def test_append_method_appends_data(self):
        data_presenter = Data_Presenter.get_instance()
        test_data_1 = sent_package("1")
        test_data_2 = receive_package("1", "1")

        data_presenter.append(test_data_1)
        data_presenter.append(test_data_2)

        latest = data_presenter.latest_packages().pop()
        assert latest.sent_at == test_data_1.time
        assert latest.received_at == test_data_2.time

    @freeze_time(auto_tick_seconds=1)
    def test_get_latest_data_by_amount_of_entries(self):
        data_presenter = Data_Presenter.get_instance()
        test_data_1 = sent_package("1")
        test_data_2 = sent_package("2")

        data_presenter.append(test_data_1)
        data_presenter.append(test_data_2)

        latest_package = data_presenter.latest_packages(1).pop()

        assert latest_package.sent_at == test_data_2.time

    def test_get_package_info_at_timestamp(self):
        data_presenter = Data_Presenter.get_instance()
        time = datetime.now()
        data_presenter.append(sent_package("1", time))
        data_presenter.append(sent_package("2", time))

        # get time table
        package_info = data_presenter.get_time_table()
        assert package_info[time].sent == 2



from steinloss.probe import Probe


class TestProbe:

    def test_packet_received_should_increment_id(self, mocker):
        mocker.patch('socket.socket')
        probe = Probe(('fake_address', 1337))

        assert probe.id == 0
        probe.receive_packet()
        assert probe.id == 1

    def test_upon_receiving_packet_should_respond_to_server_with_concat_id(self, mocker):
        mocker.patch('socket.socket')
        probe = Probe(('fake_address', 1337))

        probe.sock.recv.return_value = '1'.encode()
        probe.receive_packet()

        probe.sock.sendto.assert_called_once_with('1_0'.encode(), probe.server_address)

    def test_is_packet_loss_should_detect_loss_when_the_next_id_is_more_than_three_numbers_higere(self, mocker):
        mocker.patch('socket.socket')
        probe = Probe(('fake_address', 1337))

        probe.consume_packet('0')
        probe.consume_packet('4')
        # lost package 1
        # package 2,3 is still in reordering window

        assert probe.lost == 1

    def test_iterate_through_window_twice_without_loss(self, mocker):
        mocker.patch('socket.socket')
        probe = Probe(('fake_address', 1337))
        window_size = 5

        for n in range(0, window_size * 2):
            probe.consume_packet(str(n))

        assert probe.lost == 0

    def test_is_packet_loss_should_handle_reordering(self, mocker):
        mocker.patch('socket.socket')
        probe = Probe(('fake_address', 1337))

        probe.consume_packet('0')
        probe.consume_packet('2')
        probe.consume_packet('1')

        assert probe.lost == 0

    def test_is_packet_loss_should_handle_reordering_down_to_three(self, mocker):
        mocker.patch('socket.socket')
        probe = Probe(('fake_address', 1337))

        probe.consume_packet('0')
        probe.consume_packet('2')
        probe.consume_packet('3')
        probe.consume_packet('4')

        # not in reorder window
        probe.consume_packet('1')

        assert probe.lost == 1

    def test_no_packet_loss_inside_window(self, mocker):
        mocker.patch('socket.socket')
        probe = Probe(('fake_address', 1337))

        probe.consume_packet('1')
        probe.consume_packet('2')

        assert probe.lost == 0

    def test_detect_loss(self, mocker):
        mocker.patch('socket.socket')
        probe = Probe(('fake_address', 1337))

        probe.consume_packet('1')
        probe.consume_packet('0')
        probe.consume_packet('2')
        probe.consume_packet('4')
        probe.consume_packet('5')

        assert probe.lost == 1

# test sequence number start at first package



from datetime import datetime


class Packet_entity:
    def __init__(self, packet_content, time=datetime.now()):
        self.id = packet_content
        self.time = time

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.id == other.id
        return False

    def __repr__(self):
        return f"id: {self.id} | time:  {self.time.strftime('%H:%M:%S')}"


class sent_package(Packet_entity):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class receive_package(Packet_entity):
    def __init__(self, sent_id, recv_id, *kwargs):
        self.recv_id = recv_id
        super().__init__(sent_id, *kwargs)

    def __repr__(self):
        return f"id:{self.id} recv:{self.recv_id} time:{self.time}"



import socket

sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  # For UDP

udp_host = socket.gethostname()  # Host IP
udp_port = 5050  # specified port to connect

# print type(sock) ============> 'type' can be used to see type
# of any variable ('sock' here)
print(udp_host)

sock.bind(('', udp_port))

while True:
    print("Waiting for client...")
    data, addr = sock.recvfrom(1024)  # receive data from client
    print("Received Messages:", data, " from", addr)



from steinloss.loss_calculator import Loss_Calculator, PacketEntry, TimeTable


class Data_Presenter(object):
    __instance__ = None

    def __init__(self):
        if Data_Presenter.__instance__ is not None:
            raise RuntimeError("Cannot init class twice, as it is as singelton")
        else:
            self.data = Loss_Calculator()
            Data_Presenter.__instance__ = self

    @classmethod
    def get_instance(cls):
        """
        :rtype: Data_Presenter
        """
        if cls.__instance__ is None:
            Data_Presenter()
        return cls.__instance__

    @classmethod
    def clear_instance(cls):
        cls.__instance__ = None

    def append(self, packet):
        self.data.add(packet)

    def latest_packages(self, amount=1) -> [PacketEntry]:
        return self.data.get_last_packages(amount)

    def get_time_table(self) -> TimeTable:
        return self.data.time_table



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

REQUIREMENTS = [i.strip() for i in open("requirements.txt").readlines()]

setuptools.setup(
    name="steinloss",  # Replace with your own username
    version="0.0.2",
    author="Steinwurf ApS",
    author_email="contact@steinwurf.com",
    description="Package loss measuring tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/steinwurf/steinloss",
    entry_points={"console_scripts": ['steinloss=steinloss.stein_parser:cli']},
    packages=setuptools.find_packages(),
    install_requires=REQUIREMENTS,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Framework :: Dash"
    ],
    python_requires='>=3.8',
)



from datetime import datetime

from steinloss.packet_entity import Packet_entity, receive_package, sent_package


class TimeEntry:
    def __init__(self):
        self.loss = 0
        self.sent = 0
        self.received = 0

    def add_packet(self, packet):
        if type(packet) is sent_package:
            self.sent += 1
            self.loss += 1
        elif type(packet) is receive_package:
            self.received += 1

    def __repr__(self):
        return f"{type(self).__name__}: loss:{self.loss} - sent_package:{self.sent} received:{self.received}"


class TimeTable:
    def __init__(self):
        self.dict = dict()

    def __getitem__(self, key) -> TimeEntry:
        if isinstance(key, datetime):
            time_key = self.convert_time_to_key(key)
        else:
            time_key = key
        if time_key not in self.dict.keys():
            self.dict[time_key] = TimeEntry()
        return self.dict[time_key]

    @staticmethod
    def convert_time_to_key(packet_time: datetime):
        return packet_time.strftime("%H:%M:%S")

    def __repr__(self):
        return f"{type(self).__name__}: {str(self.dict)}"

    def __iter__(self):
        return iter(self.dict)


class PacketEntry:
    sent_at: datetime
    received_at: datetime

    def __init__(self):
        self.sent_at = None
        self.received_at = None

    def is_sent(self):
        return self.sent_at is not None

    def is_received(self):
        return self.received_at is not None

    def __repr__(self):
        return f"{type(self).__name__}: sent:{self.sent_at} → recv:{self.received_at}"


class packet_table(dict):
    def __getitem__(self, key) -> PacketEntry:
        if key not in self.keys():
            self.__setitem__(key, PacketEntry())
        return super().__getitem__(key)

    def __repr__(self):
        return f"{type(self).__name__}: {super().__repr__()}"

    def __iter__(self):
        return super().__iter__()


class Loss_Calculator:
    def __init__(self):
        self.time_table = TimeTable()
        self.packet_table = packet_table()

    def add(self, packet: Packet_entity):
        if type(packet) is sent_package:
            self.packet_table[packet.id].sent_at = packet.time
        elif type(packet) is receive_package:
            self.packet_table[packet.id].received_at = packet.time

            sent_timestamp = self.packet_table[packet.id].sent_at
            self.time_table[sent_timestamp].loss -= 1

        self.time_table[packet.time].add_packet(packet)

    def __contains__(self, item):
        if isinstance(item, Packet_entity):
            return item.id in self.packet_table
        elif isinstance(item, datetime):
            return item in self.time_table

    def __getitem__(self, key):
        if isinstance(key, datetime):
            return self.time_table[key]
        else:
            return self.packet_table[key]

    def __repr__(self):
        return f"{str(self.time_table)}\n{str(self.packet_table)}"

    def get_last_packages(self, number_of_packages: int) -> [PacketEntry]:
        arr = []
        counter = number_of_packages
        i = iter(reversed(self.packet_table))

        while counter > 0:
            key = next(i)
            arr.append(self.packet_table[key])
            counter -= 1

        return arr
